#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>

#ifdef _WIN32
#include <conio.h>
#endif

#define SM64_Z64_FNV1		0x78D3BA69F2A08D6FULL
#define SM64_Z64_FNV1A		0x623B80DDBD00E7B7ULL

#define SM64_V64_FNV		0xF3B2720E6335B7D3ULL
#define SM64_V64_FNV1A		0xFBDF5C50BDCE861BULL

#define FNV_PRIME			1099511628211ULL
#define FNV_OFFSET_BASIS	14695981039346656037ULL

#define E_OK	0
#define E_MEM	1
#define E_FILE	2

int verify_rom(const char *filename);
int v64_to_z64(const char *old, const char *new);

char *read_line(void);
char *read_filename(void);
int yes_or_no(void);
char get_option(char min, char max);

int file_exists(const char *filename);

int check_extension(const char *filename, const char *ext)
{
	size_t pos = strlen(filename);
	filename += (pos - strlen(ext) - 1);

	return *filename++ == '.' && !strcasecmp(filename, ext);
}

char *replace_extension(const char *filename, const char *ext)
{
	const char *pos = strrchr(filename, '.');
	size_t len = strlen(filename);
	size_t extlen = strlen(ext);

	// Replace the extension instead of appending it
	if (pos && pos != filename && pos > strrchr(filename, '\\'))
		len = pos - filename;

	// Append or replace extension
	char *buf = malloc(len + extlen + 2);
	if (!buf) return 0;

	strcpy(buf, filename);
	strcpy(buf + 1 + len, ext);

	buf[len] = '.';
	buf[len + 1 + extlen] = 0;

	return buf;
}

int main(int argc, const char* argv[])
{
	printf("Super Mario 64 Clean ROM Verifier v1.0, by ShadowFan-X\n");

	int e;
	if (argc > 1)
	{
		int i;
		// Verify each ROM specified in the command line
		for (i = 1; i < argc; i++)
		{
			if ((e = verify_rom(argv[i])) != 0)
				break;
			putchar('\n');
		}
	}
	else
	{
		// TODO: Prompt the user for a path.
		printf("Enter ROM name: ");
		char *filename = read_filename();
		if (!filename) return
			printf("Error: Out of memory.\n"), E_MEM;

		e = verify_rom(filename);
		free(filename);
		putchar('\n');
	}

#if _WIN32
	printf("Press any key to continue... ");
	_getch();
#else
	printf("Press Enter to continue...\n");
	getchar();
#endif
	return e;
}

// Open the ROM and check its checksum.
int verify_rom(const char *filename)
{
	printf("Verifying '%s'...\n", filename);
	FILE *pFile = fopen(filename, "rb");
	if (!pFile) return
		printf("Error: Cannot open '%s'\n", filename),
		E_FILE;

	// Clean SM64 ROM is exactly 8MB.
	fseek(pFile, 0, SEEK_END);
	if (ftell(pFile) != 0x00800000)
		goto NotClean;
	fseek(pFile, 0, SEEK_SET);

	uint64_t fnv1	= FNV_OFFSET_BASIS;
	uint64_t fnv1a	= FNV_OFFSET_BASIS;
	int c;

	while ((c = fgetc(pFile)) != EOF)
	{
		// Limit c to one byte.
		c &= 0xFF;
		// Calculate the checksums
		fnv1a	^= c;
#ifdef __GNUC__
		fnv1 += (fnv1 << 1) + (fnv1 << 4) + (fnv1 << 5) +
				(fnv1 << 7) + (fnv1 << 8) + (fnv1 << 40);
		fnv1a += (fnv1a << 1) + (fnv1a << 4) + (fnv1a << 5) +
				(fnv1a << 7) + (fnv1a << 8) + (fnv1a << 40);
#else
		fnv1	*= FNV_PRIME;
		fnv1a	*= FNV_PRIME;
#endif
		fnv1	^= c;
	}

	fclose(pFile);

	if (fnv1 == SM64_Z64_FNV1 && fnv1a == SM64_Z64_FNV1A)
	{
		printf("This is a clean Super Mario 64 ROM.\n"
			   "The ROM's format is Z64 (big-endian, AA BB CC DD)\n");
		// Check for bad extension
		if (check_extension(filename, "v64"))
		{
			printf(
				"Your ROM has the incorrect extension 'v64'.\n"
				"Do you want to fix it? (Y/N) ");
			if (yes_or_no())
			{
				char *new_filename = replace_extension(filename, "z64");
				if (!new_filename) return
					printf("Error: Out of memory.\n"),
					E_MEM;

				if (file_exists(new_filename))
				{
					printf("Error: Cannot change the extension of the ROM, "
						"because a ROM with the same name already exists.\n");
					free(new_filename);
					return E_FILE;
				}

				if (rename(filename, new_filename))
				{
					printf("Error: Cannot rename the ROM.\n");
					free(new_filename);
					return E_FILE;
				}

				free(new_filename);
			}
		}
		return E_OK;
	}
	else if (fnv1 == SM64_V64_FNV && fnv1a == SM64_V64_FNV1A)
	{
		printf("This is a clean Super Mario 64 ROM.\n"
			   "The ROM's format is V64 (PDP-endian, BB AA DD CC)\n\n");

		// Warn if there is a bad extension
		if (check_extension(filename, "z64"))
		{
			printf(
				"Your ROM has incorrect extension 'z64'.\n"
				"What do you want to do?\n"
				"\n"
				"	1. - Convert to Z64 ROM\n"
				"	2. - Fix extension\n"
				"	3. - Do nothing\n"
				"-: ");

			switch (get_option('1', '3'))
			{
			case '1':
				return v64_to_z64(filename, filename);
			case '2':; // why the fuck does that need the ';'
				char *new_filename = replace_extension(filename, "v64");
				if (!new_filename) return
					printf("Error: Out of memory.\n"),
					E_MEM;

				if (file_exists(new_filename))
				{
					printf("Error: Cannot change the extension of the ROM, "
						"because a ROM with the same name already exists.\n");
					free(new_filename);
					return E_FILE;
				}

				if (rename(filename, new_filename))
				{
					printf("Error: Cannot rename the ROM.\n");
					free(new_filename);
					return E_FILE;
				}

				free(new_filename);
				break;
			//case '3':
			default:
				break;
			}
		}
		else
		{
			printf("Do you want to convert this ROM to Z64 format? (Y/N) ");

			if (yes_or_no())
			{
				char *new_filename = replace_extension(filename, "z64");
				if (!new_filename) return
					printf("Error: Out of memory.\n"),
					E_MEM;


				if (file_exists(new_filename))
				{
					printf("The ROM already exists:\n"
						   "'%s'\n"
						   "Do you want to overwrite it? (Y/N) ", new_filename);
					if (!yes_or_no())
						return E_OK;
				}

				int n = v64_to_z64(filename, new_filename);

				free(new_filename);

				return n;
			}
		}

		return E_OK;
	}

NotClean:
	printf("This is not a clean (U) [!] Z64 or V64 ROM of Super Mario 64.\n");
	return E_OK;
}

// Convert a V64 ROM to a Z64 ROM.
int v64_to_z64(const char *old, const char *new)
{
	FILE *src, *dst;

	src = fopen(old, "rb");
	dst = fopen(new, new != old ? "wb" : "r+b");

	if (!src || !dst)
	{
		printf("Error: Cannot open the source or destination file.\n");
		if (src) fclose(src);
		if (dst) fclose(dst);
		return E_FILE;
	}

	while (1)
	{
		int a = fgetc(src);
		int b = fgetc(src);

		if (a == EOF) break;

		fputc(b, dst);
		fputc(a, dst);

		if (b == EOF) break;
	}

	fclose(src);
	if (dst != src)
		fclose(dst);

	return E_OK;
}

char *read_line(void)
{
	char * line = malloc(64), * linep = line;
	size_t lenmax = 64, len = lenmax;
	int c;

	if(!line) return 0;

	while (1)
	{
		c = fgetc(stdin);
		if(c == EOF)
			break;

		if(--len == 0)
		{
			char * linen = realloc(linep, lenmax *= 2);
			len = lenmax;

			if(linen == NULL)
			{
				free(linep);
				return NULL;
			}
			line = linen + (line - linep);
			linep = linen;
		}

		if((*line++ = c) == '\n')
			break;
    }
    // Remove the last character
    if (line[-1] == '\n')
		line--;
    *line = '\0';
    return linep;
}

char *read_filename(void)
{
	char *fn = read_line();
#ifdef _WIN32
	if (*fn == '"')
	{
		char *end = fn + strlen(fn);
		if (end[-1] == '"')
		{
			end[-1] = 0;
			fn++;
		}
	}
#endif
	return fn;
}

int yes_or_no(void)
{
#ifdef _WIN32
	while (1)
	{
		switch (_getch())
		{
		case 'Y': case 'y':
			printf("Y\n");
			return 1;
		case 'N': case 'n':
			printf("n\n");
			return 0;
		}
	}
#else
	switch (getchar())
	{
	case 'Y':
	case 'y':
		return 1;
	default:
		return 0;
	}
#endif
}

char get_option(char min, char max)
{
#ifdef _WIN32
	while (1)
	{
		char c = _getch();
		if (c >= min && c <= max)
			return printf("%c\n", c), c;
	}
#else
	return getchar();
#endif
}

// Simple way of checking if a file exists
int file_exists(const char *filename)
{
	FILE *file = fopen(filename, "r");
	if (file)
	{
		fclose(file);
		return 1;
	}
	return 0;
}
